// global variables
var orgName;
var params;
var userId;
var serverUrl;
var serverUrlModified;
var crmService;
var xrm;
try
{
    if (Xrm === undefined) xrm = parent.window.Xrm;
    else xrm = Xrm;
}
catch (err) { xrm = parent.window.Xrm; }

var ActOn; // the Act-On url for Activity History, Send Email, Hot Prospects and Dashboard

// supported entities
var Contact = "contact";
var ContactId = "contactid";
var Lead = "lead";
var LeadId = "leadid";
var Account = "account";
var AccountId = "accountid";
var Invoice = "invoice";
var InvoiceId = "invoiceid";
var SalesOrder = "salesorder";
var SalesOrderId = "salesorderid";
var Quote = "quote";
var QuoteId = "quoteid";
var Task = "task";
var TaskId = "activityid";

function EnableButton(entityName)
{
	return true;
	
	/*
	var value = false;
	if (entityName === Contact || entityName === Lead || entityName === Account)
		value = true;
	
	return value;
	*/
}

//========================================================================================================
// Initialize global variables.
//========================================================================================================
function Init()
{
	orgName = xrm.Page.context.getOrgUniqueName();
	params = xrm.Page.context.getQueryStringParameters();
	userId = xrm.Page.context.getUserId().replace( "{", "" ).replace( "}", "" );
	serverUrl = xrm.Page.context.getClientUrl();
	if (serverUrl.match(/\/$/))
	{ 
		serverUrlModified = serverUrl.substring(0, serverUrl.length - 1);
	}
	crmService = new FetchUtil(orgName, serverUrl);
}

//========================================================================================================
// Retrieve the Single-Sign-On data for the specified user.
//========================================================================================================
function fetchSSO(userId)
{
	var fetchSid = "<fetch mapping='logical' count='1'>" +
						"<entity name='custom_acton_sso'>" +
							"<attribute name='custom_sid' />" +
							"<attribute name='custom_username' />" +
							"<attribute name='custom_aologin' />" +
							"<attribute name='custom_ssoextid' />" +
							"<attribute name='custom_firstname' />" +
							"<attribute name='custom_lastname' />" +
							"<attribute name='custom_subdomain' />" +
							"<filter type='and'>" +
								"<condition attribute = 'custom_userid' operator='eq' value='" + userId + "'/>" +
							"</filter>" +
						"</entity>" +
					"</fetch>";

	var result = crmService.Fetch(fetchSid);
	
	var sid = "";
	var ssoId = "";
	var ssoExtId = "";
	var userName = "";
	var aoLogin = "";
	var firstName = "";
	var lastName = "";
	var subdomain = "";
	try { sid = (result[0].attributes["custom_sid"].value); } catch(err) {}
	try { ssoId = (result[0].attributes["custom_acton_ssoid"].value); } catch(err) {}
	try { ssoExtId = (result[0].attributes["custom_ssoextid"].value); } catch(err) {}
	try { userName = (result[0].attributes["custom_username"].value); } catch(err) {}
	try { aoLogin = (result[0].attributes["custom_aologin"].value); } catch(err) {}
	try { firstName = (result[0].attributes["custom_firstname"].value); } catch(err) {}
	try { lastName = (result[0].attributes["custom_lastname"].value); } catch(err) {}
	try { subdomain = (result[0].attributes["custom_subdomain"].value); } catch(err) {}

	ActOn = "https://" + subdomain + "/acton/"; // prod
	
	return { "sid":sid, "ssoid":ssoId, "ssoextid":ssoExtId, "username":userName, "aologin":aoLogin, "firstname":firstName, "lastname":lastName, "aoUrl":ActOn };
}

//========================================================================================================
// Get the attribute value from the Dynamics UI.
//========================================================================================================
function getAttribute(name)
{
	var attr = "";
	try
	{
		attr = xrm.Page.getAttribute(name);
		if (!isNull(attr))
			attr = attr.getValue();
	}
	catch(err) {}
	
	return attr;
}

function isNull(val) { return val === null || val === undefined; }
function isNullOrEmpty(val) { return val === "" || isNull(val); }
function notNull(value) { return isNull(value) ? "" : value; }

//========================================================================================================
// Retrieve data of the specified record (from UI and server if needed).
//========================================================================================================
function getRecordData(entity, entityId, recordId)
{
	var firstName = null;
	var lastName = null;
	
	if (entity === Account)
		firstName = getAttribute("name"); // the Account entity has the 'name' attribute
	else
		firstName = getAttribute("firstname");
		
	if (isNull(firstName))
	{
		var fullName = getAttribute("fullname");
		if (fullName !== null)
		{
			var names = fullName.split(" ");
			if (names.length > 0)
				firstName = names[0];
			if (names.length > 1)
				lastName = names[1];
		}
	}
	
	if (lastName === null)
	{
		lastName = getAttribute("lastname");
	}
	
	var email = getAttribute("emailaddress1");
	var title = getAttribute("jobtitle");
	var company = getAttribute("companyname"); // contact does not have the 'companyname' attribute
	var salutation = getAttribute("salutation");

    // alert("getRecordData: email: " + email + ", firstname: " + firstName + ", lastName: " + lastName + ", companyname: " + company);
	if (email === null) // if the email address is not available in the UI, get values from the server
		return fetchRecord(entity, entityId, recordId);
	else
		return { "fn":notNull(firstName), "ln":notNull(lastName), "email":notNull(email), "title":notNull(title), "company":notNull(company), "salutation":notNull(salutation) };
}

//========================================================================================================
// Fetch XML link entity.
//========================================================================================================
function getLinkEntity(entity, joinField)
{
	var linked = "";
	var linkedId = entity + '.' + 'linkedid';
	switch (entity)
	{
		case Contact:
			linked = "<link-entity name='contact' from='contactid' to='" + joinField + "' link-type='outer'>" +
						"<attribute name='firstname' alias='firstname' />" +
						"<attribute name='lastname' alias='lastname' />" +
						"<attribute name='emailaddress1' alias='contact.email' />" +
						"<attribute name='jobtitle' alias='jobtitle' />" +
						"<attribute name='salutation' alias='salutation' />" +
						"<attribute name='contactid' alias='" + linkedId + "' />" +
					"</link-entity>";
			break;
		case Lead:
			linked = "<link-entity name='lead' from='contactid' to='" + joinField + "' link-type='outer'>" +
						"<attribute name='firstname' alias='firstname' />" +
						"<attribute name='lastname' alias='lastname' />" +
						"<attribute name='emailaddress1' alias='emailaddress1' />" +
						"<attribute name='jobtitle' alias='jobtitle' />" +
						"<attribute name='companyname' alias='companyname' />" +
						"<attribute name='salutation' alias='salutation' />" +
					"</link-entity>";
			break;
		case Account:
			linked = "<link-entity name='account' from='accountid' to='" + joinField + "' link-type='outer'>" +
						"<attribute name='name' alias='name' />" +
						"<attribute name='emailaddress1' alias='account.email' />" +
						"<attribute name='accountid' alias='" + linkedId + "' />" +
						//"<attribute name='accountid' alias='linkedid' />" +
					"</link-entity>";
			break;
		default:
			break;
	}

	return linked;
}

//========================================================================================================
// Fetch specified entity data from Dynamics server.
//========================================================================================================
function fetchRecord(entity, entityId, recordId)
{
	var idField = EntityIdFields[entity];

	//alert('first name: ' + xrm.Page.getAttribute("firstname").getValue() + ', last name: ' + xrm.Page.getAttribute("firstname").getValue() + ', email: ' + xrm.Page.getAttribute("emailaddress1").getValue());
	var fetchXml = "<fetch mapping='logical' count='1'>" +
						"<entity name=" + "'" + entity + "'>" +
							"<attribute name='firstname' />" +
							"<attribute name='lastname' />" +
							"<attribute name='emailaddress1' />" +
							"<attribute name='jobtitle' />" +
							"<attribute name='salutation' />" +
							"<filter type='and'>" +
								"<condition attribute='" + idField + "' operator='eq' value='" + recordId + "'/>" +
							"</filter>" +
						"</entity>" +
					"</fetch>";

	switch (entity)
	{
		case Lead:
			fetchXml = "<fetch mapping='logical' count='1'>" +
							"<entity name=" + "'" + entity + "'>" +
								"<attribute name='firstname' />" +
								"<attribute name='lastname' />" +
								"<attribute name='emailaddress1' />" +
								"<attribute name='jobtitle' />" +
								"<attribute name='salutation' />" +
								"<attribute name='companyname' />" + // lead has the company field
								"<filter type='and'>" +
									"<condition attribute='" + idField + "' operator='eq' value='" + recordId + "'/>" +
								"</filter>" +
							"</entity>" +
						"</fetch>";
			break;
		case Account:
			fetchXml = "<fetch mapping='logical' count='1'>" +
							"<entity name=" + "'" + entity + "'>" +
								"<attribute name='name' />" +
								"<attribute name='emailaddress1' />" +
								"<filter type='and'>" +
									"<condition attribute='" + idField + "' operator='eq' value='" + recordId + "'/>" +
								"</filter>" +
							"</entity>" +
						"</fetch>";
			break;
		case Quote:
		case Invoice:
		case SalesOrder:
			fetchXml = "<fetch mapping='logical' count='1'>" +
							"<entity name=" + "'" + entity + "'>" +
								"<attribute name='customerid' />" +
								"<filter type='and'>" +
									"<condition attribute='" + entityId + "' operator='eq' value='" + recordId + "'/>" +
								"</filter>" + getLinkEntity(Account, 'customerid') + getLinkEntity(Contact, 'customerid') + // the customerid field which is an account (or contact) record
							"</entity>" +
						"</fetch>";
			break;
		case Task:
			fetchXml = "<fetch mapping='logical' count='1'>" +
							"<entity name=" + "'" + entity + "'>" +
								"<attribute name='regardingobjectid' />" + // regardingobject can be contact, lead or account
								"<filter type='and'>" +
									"<condition attribute='" + entityId + "' operator='eq' value='" + recordId + "'/>" +
								"</filter>" + getLinkEntity(Contact, 'regardingobjectid') +
							"</entity>" +
						"</fetch>";
			break;
	}

    var result = crmService.Fetch(fetchXml);

	var firstname = "";
	var lastname = "";
	var email = "";
	var title = "";
	var company = "";

    var attributes = result[0].attributes;
    try { firstname = (attributes.firstname.value); } catch(err) {}
	if (isNullOrEmpty(firstname))
	{
		try { firstname = (result[0].attributes["name"].value); } catch(err) {} // account name
	}
	
	try { lastname = (result[0].attributes["lastname"].value); } catch(err) {}
	
	try { email = (result[0].attributes["emailaddress1"].value); } catch(err) {}
	if (isNullOrEmpty(email))
	{
		try { email = (result[0].attributes["contact.email"].value); } catch(err) {} // contact email
		if (isNullOrEmpty(email))
			try { email = (result[0].attributes["account.email"].value); } catch(err) {}  // account email
	}
		
	try { title = (result[0].attributes["jobtitle"].value); } catch(err) {}
	try { company = (result[0].attributes["companyname"].value); } catch(err) {}

    var regardingObjectId = "";
	if (entity === Quote || entity === Invoice || entity === SalesOrder)
    {
	    try { regardingObjectId = (result[0].attributes["contact.linkedid"].value); } catch(err) {}
	    if (isNullOrEmpty(regardingObjectId))
        {
	        try { regardingObjectId = (result[0].attributes["account.linkedid"].value); } catch(err) {}
	        if (!isNullOrEmpty(regardingObjectId))
	            entity = Account;
        }
        else
        {
            entity = Contact;
        }
    }

    //alert("fn: " + firstname + ", ln: " + lastname + ", email: " + email + ", title: " + title + ", company: " + company + ", regardingObjectId: " + regardingObjectId + ", entity: " + entity);

	return { "fn":notNull(firstname), "ln":notNull(lastname), "email":notNull(email), "title":notNull(title), "company":notNull(company), "regardingobjectid":regardingObjectId, "entity":entity };
}

//========================================================================================================
// the supported entities and id fields
//========================================================================================================
var EntityIdFields = { "contact": ContactId, "lead": LeadId, "account": AccountId, "invoice": InvoiceId, "quote": QuoteId, "salesorder": SalesOrderId, "task": TaskId };

//========================================================================================================
// Activity History callback function.
//========================================================================================================
function OnActivityHistory(entityName, firstPrimaryItemId) { ActivityHistory(entityName, EntityIdFields[entityName], firstPrimaryItemId); }

//========================================================================================================
// Display the activity history for the specified record.
//========================================================================================================
function ActivityHistory(entity, entityId, recordId)
{
	Init();
	var ssoData = fetchSSO(userId);
	var record = getRecordData(entity, entityId, recordId);

    // for entities other than contact/lead/account (ex., quote, invoice,...), the entity is changed to its customer
    if (!isNullOrEmpty(record.entity))
        entity = record.entity;

    var recId = isNullOrEmpty(record.regardingobjectid) ? recordId : record.regardingobjectid;
	recId = recId.replace( "{", "" ).replace( "}", "" ).toLowerCase();

	var param = "email=" + escape(record.email) +
				"&recId=" + escape(recId) +
				"&module=" + escape(entity.toLowerCase()) +
				"&extkey=" + escape(ssoData.ssoid) +
				"&extkey1=" + escape(ssoData.ssoextid) +
				"&sid=" + escape(ssoData.sid) +
				"&user=" + escape(ssoData.aologin) +
				"&org=" + escape(orgName);
				
	// don't send empty names
	var fn = record.fn;
	if (!isNull(fn))
		param = param + "&fn=" + encodeURI(fn);
	
	var ln = record.ln;
	if (!isNull(ln))
		param = param + "&ln=" + encodeURI(ln);
	
	window.open(ActOn + "msdynamics/extActivityHistory.jsp?" + param);
}

function LeadActivityHistory(firstPrimaryItemId) { ActivityHistory(Lead, LeadId, firstPrimaryItemId); }
function ContactActivityHistory(firstPrimaryItemId) { ActivityHistory(Contact, ContactId, firstPrimaryItemId); }
function AccountActivityHistory(firstPrimaryItemId) { ActivityHistory(Account, AccountId, firstPrimaryItemId); }

//========================================================================================================
// Send email to a single record.
//========================================================================================================
function SendEmail(entity, entityId, recordId)
{	
	Init();
	var ssoData = fetchSSO(userId);
	var rec = getRecordData(entity, entityId, recordId);
		
    // for entities other than contact/lead/account (ex., quote, invoice,...), the entity is changed to its customer
    if (!isNullOrEmpty(rec.entity))
        entity = rec.entity;

    var recId = isNullOrEmpty(rec.regardingobjectid) ? recordId : rec.regardingobjectid;
	recId = recId.replace( "{", "" ).replace( "}", "" ).toLowerCase();

	var record = { 'id':notNull(recId), 'EMailAddress1':notNull(rec.email), 'JobTitle':notNull(rec.title), 'CompanyName':notNull(rec.company), 'FirstName':notNull(rec.fn), 'LastName':notNull(rec.ln), 'Salutation':notNull(rec.salutation) };
	var data = { "sid":notNull(ssoData.sid), "extkey":notNull(ssoData.ssoid), "module":notNull(entity.toLowerCase()), "extkey1":notNull(ssoData.ssoextid), "user":notNull(ssoData.aologin), "records" : [record] };
	openWindowWithPost(ActOn + "msdynamics/extSendEmail.jsp", "extSendEmails", "data", JSON.stringify(data));
}

var CrmLeadColumns = 	[ LeadId,    'emailaddress1', 'firstname', 'lastname', 'jobtitle', 'companyname', 'salutation' ];
var CrmContactColumns = [ ContactId, 'emailaddress1', 'firstname', 'lastname', 'jobtitle', 'salutation' ];
var CrmAccountColumns = [ AccountId, 'emailaddress1', 'name', 'accountnumber' ];
var LeadColumns =       [ 'id',      'EMailAddress1', 'FirstName', 'LastName', 'JobTitle', 'CompanyName', 'Salutation' ];
var ContactColumns = 	[ 'id',      'EMailAddress1', 'FirstName', 'LastName', 'JobTitle', 'Salutation' ];
var AccountColumns = 	[ 'id',      'email',         'name',      'accountnumber' ];

var contactFetchXml = "<fetch mapping='logical'> <entity name='_entity_'> <attribute name='emailaddress1' /> <attribute name='firstname' /> <attribute name='lastname' /> <attribute name='salutation' /> <attribute name='jobtitle' /> <filter type='and'> <condition attribute = '_entityId_' operator='in'> _values_</condition> </filter> </entity> </fetch>";
var leadFetchXml = "<fetch mapping='logical'> <entity name='_entity_'> <attribute name='emailaddress1' /> <attribute name='firstname' /> <attribute name='lastname' /> <attribute name='salutation' /> <attribute name='jobtitle' /> <attribute name='companyname' /> <filter type='and'> <condition attribute = '_entityId_' operator='in'> _values_</condition> </filter> </entity> </fetch>";
var accountFetchXml = "<fetch mapping='logical'> <entity name='_entity_'> <attribute name='emailaddress1' /> <attribute name='name' /> <attribute name='accountnumber' /> <filter type='and'> <condition attribute = '_entityId_' operator='in'> _values_</condition> </filter> </entity> </fetch>";

//========================================================================================================
// Send email to multiple records
//========================================================================================================
function SendEmails(entity, entityId, itemCount, itemIds, itemRefs)
{	
	if (itemCount < 1) // no record(s) selected
	{
		alert("Please select one or more " + entity + "s before performing this action.");
		return;
	}
	
	Init();
	var ssoData = fetchSSO(userId);
								
	var values = "";
	for(var i = 0; i < itemIds.length; i++)
	{
		var id = "";
		id = itemIds[i];
		values += "<value>" + id.replace( "{", "" ).replace( "}", "" ) + "</value>";
	}
		
	var isContact = (entity === Contact);
	var isLead = (entity === Lead);
	var isAccount = (entity === Account);
	var fetchXml = isContact ? contactFetchXml : isLead ? leadFetchXml : accountFetchXml;
	var fetchEmails = fetchXml.replace("_entity_", entity).replace("_entityId_", entityId).replace("_values_", values);
	var entities = crmService.Fetch(fetchEmails);
	
	var crmColumns = isContact ? CrmContactColumns : isLead ? CrmLeadColumns : CrmAccountColumns;	// lead has the company column
	var aoColumns =  isContact ? ContactColumns : LeadColumns;
	var records = RecordsToJson(entities, crmColumns, aoColumns);
	
	var data = { "sid":notNull(ssoData["sid"]), "extkey":notNull(ssoData["ssoid"]), "module":notNull(entity.toLowerCase()), "extkey1":notNull(ssoData["ssoextid"]), "user":notNull(ssoData["aologin"]), "records" : records };
	//alert("data: " + JSON.stringify(records));

	//var xmlhttp = new XMLHttpRequest();
	//xmlhttp.open("POST", ActOn + "msdynamics/extSendEmail.jsp?", false); // synchronous
	//xmlhttp.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
	//xmlhttp.send("data=" + msg);
	
	openWindowWithPost(ActOn + "msdynamics/extSendEmail.jsp?", "extSendEmails", "data", JSON.stringify(data));
	
	//var myWindow = window.open("", "", "height=600,width=800,scrollbars=1,location=no,menubar=no,resizable=1,status=no,toolbar=no");
	//var reply = xmlhttp.send("data=" + msg);
	//alert("reply: " + reply);
	//myWindow.document.write(reply);
 }

//========================================================================================================
// Called when the Send Emails button in Homepage is clicked.
//========================================================================================================
function OnSendEmails(entityName, itemCount, itemIds, itemRefs)
{
	if (entityName === Contact)
		ContactSendEmails(itemCount, itemIds, itemRefs);
	else if (entityName === Lead)
		LeadSendEmails(itemCount, itemIds, itemRefs);
	else if (entityName === Account)
		AccountSendEmails(itemCount, itemIds, itemRefs);
}

//========================================================================================================
// Called when the Send Email button in a form is clicked.
//========================================================================================================
function OnSendEmail(entityName, firstPrimaryItemId)
{
	SendEmail(entityName, EntityIdFields[entityName], firstPrimaryItemId);
}

function SyncToActOnCallback(data)
{
	if (data.status === 'ok')
		alert('Sync completed.');
	else if (!isNull(data.message ))
		alert('Sync failed, error: ' + data.message);
}

// load jQuery if not loaded yet (Syn to Act-On)
if (!window.jQuery)
{
	var jQuery213 = document.createElement('script');
	//jQuery213.src = serverUrl + '//WebResources/custom_jquery.min.2.1.3.js';
	jQuery213.src = 'https://code.jquery.com/jquery-2.1.3.min.js';
	jQuery213.type = 'text/javascript';
	document.getElementsByTagName('head')[0].appendChild(jQuery213);
}

function OnSyncToActOn(entity, firstPrimaryItemId)
{
	Init();
	var ssoData = fetchSSO(userId);
	var recordId = firstPrimaryItemId.replace( "{", "" ).replace( "}", "" ).toLowerCase();
		
	var firstName = "";
	var lastName = "";
	var email = "";
	var title = "";
	var company = "";
	var salutation = "";
	
	try { salutation = xrm.Page.getAttribute("salutation").getValue(); } catch(err) {}	
	try { firstName = xrm.Page.getAttribute("firstname").getValue(); } catch(err) {}
	try { lastName = xrm.Page.getAttribute("lastname").getValue(); } catch(err) {}
	try { email = xrm.Page.getAttribute("emailaddress1").getValue(); } catch(err) {}
	try { title = xrm.Page.getAttribute("jobtitle").getValue(); } catch(err) {}
	try { company = xrm.Page.getAttribute("companyname").getValue(); } catch(err) {}
	
	var record = { 'id':notNull(recordId), 'EMailAddress1':notNull(email), 'JobTitle':notNull(title), 'CompanyName':notNull(company), 'Salutation':notNull(salutation), 'FirstName':notNull(firstName), 'LastName':notNull(lastName) };
	var data = { "sid":notNull(ssoData["sid"]), "extkey":notNull(ssoData["ssoid"]), "module":notNull(entity.toLowerCase()), "extkey1":notNull(ssoData["ssoextid"]), "user":notNull(ssoData["aologin"]), "records" : [record] };

	//if (jQuery213 === null)		

	$.ajax({
		url: ActOn + "msdynamics/extUpdateRecord.jsp" + '?data=' + JSON.stringify(data),
		dataType: "jsonp",
		async: false,
		jsonpCallback: "SyncToActOnCallback"
	});
}

function ContactSendEmail(FirstPrimaryItemId) { SendEmail(Contact, ContactId, FirstPrimaryItemId); }
function LeadSendEmail(FirstPrimaryItemId) { SendEmail(Lead, LeadId, FirstPrimaryItemId); }
function AccountSendEmail(FirstPrimaryItemId) { SendEmail(Account, AccountId, FirstPrimaryItemId); }

function ContactSendEmails(itemCount, itemIds, itemRefs) { SendEmails(Contact, ContactId, itemCount, itemIds, itemRefs); }
function LeadSendEmails(itemCount, itemIds, itemRefs) { SendEmails(Lead, LeadId, itemCount, itemIds, itemRefs); }
function AccountSendEmails(itemCount, itemIds, itemRefs) { SendEmails(Account, AccountId, itemCount, itemIds, itemRefs); }

function openWindowWithPost(url, name, key, value)
{
	var target = "Send_Emails";
    var form = document.createElement("form");
    form.target = target;
    form.method = "post";
    form.action = url;

    var input = document.createElement("input");
    input.type = "hidden";
    input.name = key;
    input.value = value;
    form.appendChild(input);

    document.body.appendChild(form);

    var newWindow = window.open("", target);
    //map = window.open("", "Send_Emails", "status=0,title=0,height=600,width=800,scrollbars=1");

	if (newWindow)
	{
		form.submit();
	}
	else
	{
		alert('You must allow popups for Send Email to work.');
	}
}

//========================================================================================================
// Create JSON objects with all the SSO fields.
//========================================================================================================
function SsoJson()
{
	var param = "&extkey=" + escape(ssoData["ssoid"]) +
				"&extkey1=" + escape(ssoData["ssoextid"]) +
				"&sid=" + escape(ssoData["sid"]) +
				"&user=" + escape(ssoData["aologin"]) +
				"&module=" + escape(entity.toLowerCase()) +
				"&records=" + escape(JSON.stringify([{ "id":recordId, "module":entity, "email":xrm.Page.getAttribute("emailaddress1").getValue() }])); // for some reason the GUID is in upper case format that prevent looking up properly in Act-On
	//JSON.stringify([{ "ssoid":ssoData["ssoid"], "module":entity.toLowerCase(), "ssoextid":ssoData["ssoextid"], "aologin":ssoData["aologin"] }]));
}

function toJsonArray(name, entities, keys)
{
	var data = '{"' + name + '":[';
	for(var i = 0; i < entities.length; i++)
	{
		if (i > 0)
			data += ',';
			
		data += '{';
		for (var keyIndex = 0; keyIndex < keys.length; keyIndex++)
		{
			if (keyIndex > 0)
				data += ',';
			data += '"' + keys[keyIndex] + '":"' + entities[i].attributes[keys[keyIndex]].value + '"';
		}
		data += '}';
	}
	data += ']}';
	
	return data;
}

function RecordsToJson(entities, crmNames, aoNames)
{
	var records = new Array(entities.length);
	for(var i = 0; i < entities.length; i++)
	{
		var obj = {};
		for (var keyIndex = 0; keyIndex < crmNames.length; keyIndex++)
		{
			var key = crmNames[keyIndex];
			var name = aoNames[keyIndex];
			var attr = entities[i].attributes[key];
			if (isNull(attr))
				obj[name] = "";
			else
				obj[name] = entities[i].attributes[key].value; // use "id" as the key name that works for both contact and lead
		}
		records[i] = obj;
	}
	
	//alert("RecordsToJson: " + JSON.stringify(records));
	return records;
}

var XMLHTTPSUCCESS = 200;
var XMLHTTPREADY = 4;

function FetchUtil(sOrg, sServer)
{
	this.org = sOrg;
	this.server = sServer;


	if (sServer == null)
	{
		this.server = window.location.protocol + "//" + window.location.host;
	}
}

FetchUtil.prototype._ExecuteRequest = function (sXml, sMessage, fInternalCallback, fUserCallback)
{
	var xmlhttp = new XMLHttpRequest();
	xmlhttp.open("POST", this.server + "/XRMServices/2011/Organization.svc/web", (fUserCallback != null));
	xmlhttp.setRequestHeader("Accept", "application/xml, text/xml, */*");
	xmlhttp.setRequestHeader("Content-Type", "text/xml; charset=utf-8");
	xmlhttp.setRequestHeader("SOAPAction", "http://schemas.microsoft.com/xrm/2011/Contracts/Services/IOrganizationService/Execute");

	if (fUserCallback != null)
	{
		//asynchronous: register callback function, then send the request.
		var crmServiceObject = this;
		xmlhttp.onreadystatechange = function ()
		{
			fInternalCallback.call(crmServiceObject, xmlhttp, fUserCallback)
		};
		xmlhttp.send(sXml);
	}
	else
	{
		//synchronous: send request, then call the callback function directly
		xmlhttp.send(sXml);
		return fInternalCallback.call(this, xmlhttp, null);
	}
}

FetchUtil.prototype._HandleErrors = function (xmlhttp)
{
	/// <summary>(private) Handles xmlhttp errors</summary>
	if (xmlhttp.status != XMLHTTPSUCCESS)
	{
		var sError = "Error: " + xmlhttp.responseText + " " + xmlhttp.statusText;
		alert(sError);
		return true;
	}
	else
	{
		return false;
	}
}

FetchUtil.prototype.Fetch = function (sFetchXml, fCallback)
{
	/// <summary>Execute a FetchXml request. (result is the response XML)</summary>
	/// <param name="sFetchXml">fetchxml string</param>
	/// <param name="fCallback" optional="true" type="function">(Optional) Async callback function if specified. If left null, function is synchronous </param>

	var request = "<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\">";
	request += '<s:Body><Execute xmlns="http://schemas.microsoft.com/xrm/2011/Contracts/Services"><request i:type="b:RetrieveMultipleRequest" xmlns:b="http://schemas.microsoft.com/xrm/2011/Contracts" xmlns:i="http://www.w3.org/2001/XMLSchema-instance"><b:Parameters xmlns:c="http://schemas.datacontract.org/2004/07/System.Collections.Generic"><b:KeyValuePairOfstringanyType><c:key>Query</c:key><c:value i:type="b:FetchExpression"><b:Query>';
	request += parent.CrmEncodeDecode.CrmXmlEncode(sFetchXml);
	request += '</b:Query></c:value></b:KeyValuePairOfstringanyType></b:Parameters><b:RequestId i:nil="true"/><b:RequestName>RetrieveMultiple</b:RequestName></request></Execute></s:Body></s:Envelope>';

	return this._ExecuteRequest(request, "Fetch", this._FetchCallback, fCallback);
}

var isIE = navigator.userAgent.toLowerCase().indexOf("msie") != -1;
var isIE10 = false;

//========================================================================================================
// The callback function for IE (10 and older versions)
//========================================================================================================
function IECallback(xmlhttp, callback)
{
	var responseXML = null;
	try
	{
		xmlhttp.responseType = 'msxml-document';
		responseXML = xmlhttp.responseXML.selectSingleNode("//a:Entities").xml;
	}
	catch(e)
	{
		isIE10 = true;
		responseXML = new XMLSerializer().serializeToString(xmlhttp.responseXML);
	}

	//var doc = new ActiveXObject("Msxml2.DOMDocument.6.0");
	var doc = new ActiveXObject("Microsoft.XMLDOM");
	doc.async = false;
	doc.loadXML(responseXML);

	//doc.setProperty("SelectionLanguage", "XPath");
	var docEntities = doc.selectSingleNode("//a:Entities/a:Entity");
	//alert('docEntities: ' + docEntities.xml);

	//parse docEntities xml into array of jsDynamicEntity objects
	var count = docEntities.childNodes.length;
	var results = null;

	for (var i = 0; i < count; i++)
	{
		var oResultNode = docEntities.childNodes[i];
		var jDE = new jsDynamicEntity();
		var obj = new Object();

		var baseName = oResultNode.baseName;
		switch (baseName)
		{
			case "Attributes":
				results = new Array(oResultNode.childNodes.length); // create the result array for storing all attributes

				for (var k = 0; k < oResultNode.childNodes.length; k++)
				{
					var attr = oResultNode.childNodes[k];
					var entCV = new jsCrmValue();
					for (var l = 0; l < attr.childNodes.length; l++)
					{
						var attrBaseName = attr.childNodes[l].baseName;
						switch (attrBaseName)
						{
							case "key":
								entCV.key = attr.childNodes[l].text;
								break;
							case "value":
								entCV.value = attr.childNodes[l].text;
								break;
							default:
								break;
						}
					}

					//alert('entCV: ' + entCV.key + ', entCV.value: ' + entCV.value);
					obj[entCV.key] = entCV;
				}
				jDE.attributes = obj;
				break;
			case "Id":
				jDE.guid = oResultNode.text;
				//alert('jDE.guid: ' + jDE.guid);
				break;
			case "LogicalName":
				jDE.logicalName = oResultNode.text;
				//alert('LogicalName: ' + jDE.logicalName);
				break;
			case "FormattedValues":
				break;
		}
		results[i] = jDE;
	}
	
	return results;
}

//==============================================================================================================================
// The callback function for non-IE browsers (like Firefox), somehow Chrome has problem with the xrm function and does not work.
//==============================================================================================================================
function GeneralCallback (xmlhttp, callback)
{
	var parser=new DOMParser();
	var xmlDoc = parser.parseFromString(xmlhttp.responseText, "text/xml");

	// Chrome does not handle name-space well but FF does
	var entities = xmlDoc.getElementsByTagName("Entity");
	if (entities.length <= 0)
		entities = xmlDoc.getElementsByTagName("a:Entity"); // FF
		
	//var totalCount = xmlhttp.responseXML.getElementsByTagName("a:TotalRecordCount")[0].textContent;

	var results = new Array(entities.length); // create the result array for storing all attributes

	for (var i = 0; i < entities.length; i++)
	{
		var oResultNode = entities[i];

		var jDE = new jsDynamicEntity();
		var obj = new Object();

		for (var j = 0; j < oResultNode.childElementCount; j++)
		{
			var attr = oResultNode.childNodes[j];
			var baseName = attr.localName;
			switch (baseName)
			{
				case "Attributes":
					for (var k = 0; k < attr.childElementCount; k++)
					{
						var KeyValuePairOfstringanyType = attr.childNodes[k]; // KeyValuePairOfstringanyType element contains "key" and "value" elements
						var entCV = new jsCrmValue();
						for (var l = 0; l < KeyValuePairOfstringanyType.childElementCount; l++)
						{
							var childNode = KeyValuePairOfstringanyType.childNodes[l];
							switch (childNode.localName)
							{
								case "key":
									entCV.key =  childNode.childNodes[0].textContent;
									break;
								case "value":
									entCV.value = childNode.childNodes[0].textContent;
									if (childNode.hasAttributes())
									{
										var attrValue = childNode.getAttribute("i:type");
										if (attrValue.indexOf("AliasedValue") > -1) // handle the alias value
										{
											for (var m = 0; l < childNode.childElementCount; m++)
											{
												var node = childNode.childNodes[m];
												if (node.localName === "Value")
												{
													entCV.value = node.textContent;
													//alert('AliasedValue.Value: ' +  entCV.value);
													break;
												}
											}
										}
									}
									break;
								default:
									break;
							}
						}
						obj[entCV.key] = entCV;
					}

				jDE.attributes = obj;
				break;

				case "Id":
					jDE.guid = attr.text;
					//alert('jDE.guid: ' + jDE.guid);
					break;

				case "LogicalName":
					jDE.logicalName = attr.text;
					//alert('LogicalName: ' + jDE.logicalName);
					break;

				case "FormattedValues":
					break;
			}
		}
	results[i] = jDE;
	}

	return results;
}

function msieVersion()
{
	var ua = window.navigator.userAgent
	var msie = ua.indexOf ( "MSIE " )

	if ( msie > 0 )      // If Internet Explorer, return version number
	{
		var start = msie+5;
		var version = ua.substring (start, ua.indexOf (".", start));
		return parseInt (version);
	}
	else                 // If another browser, return 0
		return 0;
	}
   
FetchUtil.prototype._FetchCallback = function (xmlhttp, callback)
{
	///<summary>(private) Fetch message callback.</summary>
	if (xmlhttp.readyState != XMLHTTPREADY || this._HandleErrors(xmlhttp)) // return if the xmlhttp is not completed, or any server error
	{
		return;
	}

	var ieVersion = msieVersion();

	if (ieVersion > 0 && ieVersion < 11) // old IEs 
		return IECallback(xmlhttp, callback);
	else (window.DOMParser)
		return GeneralCallback(xmlhttp, callback);
}
 
function jsDynamicEntity(gID, sLogicalName)
{
	this.guid = gID;
	this.logicalName = sLogicalName;
	this.attributes = new Object();
}

function jsCrmValue(sType, sValue)
{
	this.type = sType;
	this.value = sValue;
}

function jsEntityReference(gID, sLogicalName, sName)
{
	this.guid = gID;
	this.logicalName = sLogicalName;
	this.name = sName;
	this.type = 'EntityReference';
}

function jsOptionSetValue(iValue, sFormattedValue)
{
	this.value = iValue;
	this.formattedValue = sFormattedValue;
	this.type = 'OptionSetValue';
}